# Copyright (c) 2015 Presslabs SRL
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

import datetime as dt
import logging

from decimal import Decimal

from django.utils import timezone

from silver.models import Customer, Subscription, Proforma, Invoice, Provider, BillingLog
from silver.utils.dates import ONE_DAY


logger = logging.getLogger(__name__)


class DocumentsGenerator(object):
    def generate(self, subscription=None, billing_date=None, customers=None,
                 force_generate=False):
        """
        The `public` method called when one wants to generate the billing documents.

        :param subscription: the subscription for which one wants to generate the
            proformas/invoices.
        :param billing_date: the date used as billing date
        :param customers: the customers for which one wants to generate the
            proformas/invoices.
        :param force_generate: if True, invoices are generated at the date
            indicated by `billing_date` instead of the normal end of billing
            cycle.

        :note
                If `subscription` is passed, only the documents for that subscription are
            generated.
                If the `customers` parameter is passed, only the docments for those customers are
            generated.
                Only one of the `customers` and `subscription` parameters may be passed at a time.
                If neither the `subscription` nor the `customers` parameters are passed, the
                documents for all the customers will be generated.
        """

        if not subscription:
            customers = customers or Customer.objects.all()
            self._generate_all(billing_date=billing_date,
                               customers=customers,
                               force_generate=force_generate)
        else:
            self._generate_for_single_subscription(subscription=subscription,
                                                   billing_date=billing_date,
                                                   force_generate=force_generate)

    def _generate_all(self, billing_date=None, customers=None, force_generate=False):
        """
        Generates the invoices/proformas for all the subscriptions that should
        be billed.
        """

        billing_date = billing_date or timezone.now().date()
        # billing_date -> the date when the billing documents are issued.

        for customer in customers:
            if customer.consolidated_billing:
                self._generate_for_user_with_consolidated_billing(
                    customer, billing_date, force_generate
                )
            else:
                self._generate_for_user_without_consolidated_billing(
                    customer, billing_date, force_generate
                )

    def _log_subscription_billing(self, document, subscription):
        logger.debug('Billing subscription: %s', {
            'subscription': subscription.id,
            'state': subscription.state,
            'doc_type': document.provider.flow,
            'number': document.number,
            'provider': document.provider.id,
            'customer': document.customer.id
        })

    def get_subscriptions_prepared_for_billing(self, customer, billing_date, force_generate):
        # Select all the active or canceled subscriptions
        subs_to_bill = []
        criteria = {'state__in': [Subscription.STATES.ACTIVE,
                                  Subscription.STATES.CANCELED]}
        for subscription in customer.subscriptions.filter(**criteria):
            if subscription.should_be_billed(billing_date) or force_generate:
                subs_to_bill.append(subscription)

        return subs_to_bill

    def _bill_subscription_into_document(self, subscription, billing_date, document=None):
        if not document:
            document = self._create_document(subscription, billing_date)

        self._log_subscription_billing(document, subscription)

        kwargs = subscription.billed_up_to_dates

        kwargs.update({
            'billing_date': billing_date,
            'subscription': subscription,
            subscription.provider.flow: document,
        })
        self.add_subscription_cycles_to_document(**kwargs)

        if subscription.state == Subscription.STATES.CANCELED:
            subscription.end()
            subscription.save()

        return document

    def _generate_for_user_with_consolidated_billing(self, customer, billing_date, force_generate):
        """
        Generates the billing documents for all the subscriptions of a customer
        who uses consolidated billing.
        """

        # For each provider there will be one invoice or proforma. The cache is necessary as a
        # certain customer might have more than one subscription
        # => all the subscriptions belonging to the same provider will be added to the same document

        existing_provider_documents = {}
        for subscription in self.get_subscriptions_prepared_for_billing(customer, billing_date,
                                                                        force_generate):
            provider = subscription.plan.provider

            existing_document = existing_provider_documents.get(provider)

            existing_provider_documents[provider] = self._bill_subscription_into_document(
                subscription, billing_date, document=existing_document
            )

        for provider, document in existing_provider_documents.items():
            if provider.default_document_state == Provider.DEFAULT_DOC_STATE.ISSUED:
                document.issue()

    def _generate_for_user_without_consolidated_billing(self, customer, billing_date,
                                                        force_generate):
        """
        Generates the billing documents for all the subscriptions of a customer
        who does not use consolidated billing.
        """

        # The user does not use consolidated_billing => add each subscription to a separate document
        for subscription in self.get_subscriptions_prepared_for_billing(customer, billing_date,
                                                                        force_generate):
            provider = subscription.plan.provider

            document = self._bill_subscription_into_document(subscription, billing_date)

            if provider.default_document_state == Provider.DEFAULT_DOC_STATE.ISSUED:
                document.issue()

    def _generate_for_single_subscription(self, subscription=None, billing_date=None,
                                          force_generate=False):
        """
        Generates the billing documents corresponding to a single subscription.
        Usually used when a subscription is ended with `when`=`now`.
        """

        billing_date = billing_date or timezone.now().date()

        provider = subscription.provider

        if not subscription.should_be_billed(billing_date) or force_generate:
            return

        document = self._bill_subscription_into_document(subscription, billing_date)

        if provider.default_document_state == Provider.DEFAULT_DOC_STATE.ISSUED:
            document.issue()

    def add_subscription_cycles_to_document(self, billing_date, metered_features_billed_up_to,
                                            plan_billed_up_to, subscription,
                                            proforma=None, invoice=None):
        relative_start_date = metered_features_billed_up_to + ONE_DAY
        plan_now_billed_up_to = plan_billed_up_to
        metered_features_now_billed_up_to = metered_features_billed_up_to

        prebill_plan = subscription.prebill_plan

        plan_amount = Decimal('0.00')
        metered_features_amount = Decimal('0.00')

        last_cycle_end_date = subscription.cycle_end_date(billing_date)

        # We iterate through each cycle (multiple bucket cycles can be contained within a billing
        # cycle) and add the entries to the document

        # relative_start_date and relative_end_date define the cycle that is billed within the
        # loop's iteration (referred throughout the comments as the cycle)
        while relative_start_date <= last_cycle_end_date:
            relative_end_date = subscription.bucket_end_date(
                reference_date=relative_start_date
            )

            if not relative_end_date:
                # There was no cycle for the given billing date
                break

            # This is here in order to separate the trial entries from the paid ones
            if (subscription.trial_end and
                    relative_start_date <= subscription.trial_end <= relative_end_date):
                relative_end_date = subscription.trial_end

            # This cycle decision, based on cancel_date, should be moved into `cycle_start_date` and
            # `cycle_end_date`
            if subscription.cancel_date:
                relative_end_date = min(subscription.cancel_date, relative_end_date)

            # If the plan is prebilled we can only bill it if the cycle hasn't been billed before;
            # If the plan is not prebilled we can only bill it if the cycle has ended before the
            # billing date.
            should_bill_plan = ((plan_billed_up_to < relative_start_date) if prebill_plan else
                                (relative_end_date < billing_date))

            # Bill the plan amount
            if should_bill_plan:
                if subscription.on_trial(relative_start_date):
                    plan_amount += subscription._add_plan_trial(start_date=relative_start_date,
                                                                end_date=relative_end_date,
                                                                invoice=invoice, proforma=proforma)
                else:
                    plan_amount += subscription._add_plan_value(start_date=relative_start_date,
                                                                end_date=relative_end_date,
                                                                proforma=proforma, invoice=invoice)
                plan_now_billed_up_to = relative_end_date

            # Only bill metered features if the cycle the metered features belong to has ended
            # before the billing date.
            should_bill_metered_features = relative_end_date < billing_date

            # Bill the metered features
            if should_bill_metered_features:
                if subscription.on_trial(relative_start_date):
                    metered_features_amount += subscription._add_mfs_for_trial(
                        start_date=relative_start_date, end_date=relative_end_date,
                        invoice=invoice, proforma=proforma
                    )
                else:
                    metered_features_amount += subscription._add_mfs(
                        start_date=relative_start_date, end_date=relative_end_date,
                        proforma=proforma, invoice=invoice
                    )

                metered_features_now_billed_up_to = relative_end_date

            # Obtain a start date for the next iteration (cycle)
            relative_start_date = relative_end_date + ONE_DAY

            if relative_end_date == subscription.cancel_date:
                break

        BillingLog.objects.create(subscription=subscription,
                                  invoice=invoice, proforma=proforma,
                                  total=plan_amount + metered_features_amount,
                                  plan_amount=plan_amount,
                                  metered_features_amount=metered_features_amount,
                                  billing_date=billing_date,
                                  metered_features_billed_up_to=metered_features_now_billed_up_to,
                                  plan_billed_up_to=plan_now_billed_up_to)

    def _create_document(self, subscription, billing_date):
        provider = subscription.provider
        customer = subscription.customer

        DocumentModel = (Proforma if provider.flow == provider.FLOWS.PROFORMA
                         else Invoice)

        payment_due_days = dt.timedelta(days=customer.payment_due_days)
        due_date = billing_date + payment_due_days
        document = DocumentModel.objects.create(provider=provider,
                                                customer=customer,
                                                due_date=due_date,
                                                currency=subscription.plan.currency)

        return document



# Copyright (c) 2016 Presslabs SRL
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import, unicode_literals

import logging
from datetime import datetime, timedelta
from decimal import Decimal

from django_fsm import FSMField, transition, TransitionNotAllowed, post_transition
from model_utils import Choices

from django.apps import apps
from django.db.models import JSONField
from django.db.models.signals import post_save
from django.dispatch import receiver
from django.conf import settings
from django.core.exceptions import ValidationError, NON_FIELD_ERRORS
from django.core.serializers.json import DjangoJSONEncoder
from django.urls import reverse
from django.core.validators import MinValueValidator
from django.db import models
from django.db import transaction as db_transaction
from django.db.models import Max, ForeignKey, F
from django.template.loader import select_template
from django.utils import timezone
from django.utils.encoding import force_str
from django.utils.safestring import mark_safe
from django.utils.text import slugify
from django.utils.translation import gettext_lazy as _
from django.utils.module_loading import import_string

from silver.currencies import CurrencyConverter, RateNotFound
from silver.models.billing_entities import Customer, Provider
from silver.models.documents.entries import DocumentEntry
from silver.models.documents.pdf import PDF
from silver.utils.decorators import require_transaction_currency_and_xe_rate
from silver.utils.international import currencies


_storage = getattr(settings, 'SILVER_DOCUMENT_STORAGE', None)
if _storage:
    _storage_klass = import_string(_storage[0])
    _storage = _storage_klass(*_storage[1], **_storage[2])

PAYMENT_DUE_DAYS = getattr(settings, 'SILVER_DEFAULT_DUE_DAYS', 5)

logger = logging.getLogger(__name__)


def documents_pdf_path(document, filename):
    path = '{prefix}{company}/{doc_name}/{date}/{filename}'.format(
        company=slugify(force_str(
            document.provider.company or document.provider.name)),
        date=document.issue_date.strftime('%Y/%m'),
        doc_name=('%ss' % document.__class__.__name__).lower(),
        prefix=getattr(settings, 'SILVER_DOCUMENT_PREFIX', ''),
        filename=filename)
    return path


class BillingDocumentQuerySet(models.QuerySet):
    def due_this_month(self):
        return self.filter(
            state=BillingDocumentBase.STATES.ISSUED,
            due_date__gte=timezone.now().date().replace(day=1)
        )

    def due_today(self):
        return self.filter(
            state=BillingDocumentBase.STATES.ISSUED,
            due_date__exact=timezone.now().date()
        )

    def overdue(self):
        return self.filter(
            state=BillingDocumentBase.STATES.ISSUED,
            due_date__lt=timezone.now().date()
        )

    def overdue_since_last_month(self):
        return self.filter(
            state=BillingDocumentBase.STATES.ISSUED,
            due_date__lt=timezone.now().date().replace(day=1)
        )


class BillingDocumentManager(models.Manager):
    def get_queryset(self):
        return super(BillingDocumentManager, self).get_queryset() \
                                                  .select_related('customer', 'provider',
                                                                  'related_document')


def get_billing_documents_kinds():
    return ((subclass.__name__.lower(), subclass.__name__)
            for subclass in BillingDocumentBase.__subclasses__())


class BillingDocumentBase(models.Model):
    objects = BillingDocumentManager.from_queryset(BillingDocumentQuerySet)()

    class STATES(object):
        DRAFT = 'draft'
        ISSUED = 'issued'
        PAID = 'paid'
        CANCELED = 'canceled'

    STATE_CHOICES = Choices(
        (STATES.DRAFT, _('Draft')),
        (STATES.ISSUED, _('Issued')),
        (STATES.PAID, _('Paid')),
        (STATES.CANCELED, _('Canceled'))
    )

    kind = models.CharField(get_billing_documents_kinds, max_length=8, db_index=True)
    related_document = models.ForeignKey('self', blank=True, null=True, on_delete=models.SET_NULL,
                                         related_name='reverse_related_document')

    series = models.CharField(max_length=20, blank=True, null=True,
                              db_index=True)
    number = models.IntegerField(blank=True, null=True, db_index=True)
    customer = models.ForeignKey('Customer', on_delete=models.CASCADE)
    provider = models.ForeignKey('Provider', on_delete=models.CASCADE)
    archived_customer = JSONField(default=dict, null=True, blank=True, encoder=DjangoJSONEncoder)
    archived_provider = JSONField(default=dict, null=True, blank=True, encoder=DjangoJSONEncoder)
    due_date = models.DateField(null=True, blank=True)
    issue_date = models.DateField(null=True, blank=True, db_index=True)
    paid_date = models.DateField(null=True, blank=True)
    cancel_date = models.DateField(null=True, blank=True)
    sales_tax_percent = models.DecimalField(max_digits=4, decimal_places=2,
                                            validators=[MinValueValidator(0.0)],
                                            null=True, blank=True)
    sales_tax_name = models.CharField(max_length=64, blank=True, null=True)
    currency = models.CharField(
        choices=currencies, max_length=4, default='USD',
        help_text='The currency used for billing.'
    )
    transaction_currency = models.CharField(
        choices=currencies, max_length=4,
        help_text='The currency used when making a transaction.'
    )
    transaction_xe_rate = models.DecimalField(
        max_digits=16, decimal_places=4, null=True, blank=True,
        help_text='Currency exchange rate from document currency to '
                  'transaction_currency.'
    )
    transaction_xe_date = models.DateField(
        null=True, blank=True,
        help_text='Date of the transaction exchange rate.'
    )

    pdf = ForeignKey(PDF, null=True, blank=True, on_delete=models.SET_NULL)
    state = FSMField(choices=STATE_CHOICES, max_length=10, default=STATES.DRAFT,
                     verbose_name="State",
                     help_text='The state the invoice is in.')

    _total = models.DecimalField(max_digits=19, decimal_places=2,
                                 null=True, blank=True)
    _total_in_transaction_currency = models.DecimalField(max_digits=19,
                                                         decimal_places=2,
                                                         null=True, blank=True)

    is_storno = models.BooleanField(default=False)

    _last_state = None
    _document_entries = None

    class Meta:
        unique_together = ('kind', 'provider', 'series', 'number')
        ordering = ('-issue_date', 'series', '-number')

    def __init__(self, *args, **kwargs):
        super(BillingDocumentBase, self).__init__(*args, **kwargs)

        if not self.kind:
            self.kind = self.__class__.__name__.lower()
        else:
            for subclass in BillingDocumentBase.__subclasses__():
                if subclass.__name__.lower() == self.kind:
                    self.__class__ = subclass

        self._last_state = self.state

    def _get_entries(self):
        if not self._document_entries:
            self._document_entries = getattr(self, self.kind + '_entries').all()

        return self._document_entries

    def compute_total_in_transaction_currency(self):
        return sum([Decimal(entry.total_in_transaction_currency)
                    for entry in self._get_entries()])

    def compute_total(self):
        return sum([Decimal(entry.total)
                    for entry in self._get_entries()])

    def mark_for_generation(self):
        self.pdf.mark_as_dirty()

    def _issue(self, issue_date=None, due_date=None):
        if issue_date:
            self.issue_date = datetime.strptime(issue_date, '%Y-%m-%d').date()
        elif not self.issue_date and not issue_date:
            self.issue_date = timezone.now().date()

        if not self.transaction_xe_rate:
            if not self.transaction_xe_date:
                self.transaction_xe_date = self.issue_date

            try:
                xe_rate = CurrencyConverter.convert(1, self.currency,
                                                    self.transaction_currency,
                                                    self.transaction_xe_date)
            except RateNotFound:
                raise TransitionNotAllowed('Couldn\'t automatically obtain an '
                                           'exchange rate.')

            self.transaction_xe_rate = xe_rate

        if not self.is_storno:
            if due_date:
                self.due_date = datetime.strptime(due_date, '%Y-%m-%d').date()
            elif not self.due_date and not due_date:
                delta = timedelta(days=PAYMENT_DUE_DAYS)
                self.due_date = timezone.now().date() + delta

        if not self.sales_tax_name:
            self.sales_tax_name = self.customer.sales_tax_name
        if not self.sales_tax_percent:
            self.sales_tax_percent = self.customer.sales_tax_percent

        if not self.number:
            self.number = self._generate_number()

        self.archived_customer = self.customer.get_archivable_field_values()
        self._total = self.compute_total()
        self._total_in_transaction_currency = self.compute_total_in_transaction_currency()

    @transition(field=state, source=STATES.DRAFT, target=STATES.ISSUED)
    def issue(self, issue_date=None, due_date=None):
        self._issue(issue_date=issue_date, due_date=due_date)

    def _pay(self, paid_date=None):
        if paid_date:
            self.paid_date = datetime.strptime(paid_date, '%Y-%m-%d').date()
        if not self.paid_date and not paid_date:
            self.paid_date = timezone.now().date()

    @transition(field=state, source=STATES.ISSUED, target=STATES.PAID)
    def pay(self, paid_date=None):
        self._pay(paid_date=paid_date)

    def _cancel(self, cancel_date=None):
        if cancel_date:
            self.cancel_date = datetime.strptime(cancel_date, '%Y-%m-%d').date()
        if not self.cancel_date and not cancel_date:
            self.cancel_date = timezone.now().date()

    @transition(field=state, source=STATES.ISSUED, target=STATES.CANCELED)
    def cancel(self, cancel_date=None):
        self._cancel(cancel_date=cancel_date)

    def sync_related_document_state(self):
        if self.is_storno:
            return

        if self.related_document and self.state != self.related_document.state:
            state_transition_map = {
                BillingDocumentBase.STATES.ISSUED: 'issue',
                BillingDocumentBase.STATES.CANCELED: 'cancel',
                BillingDocumentBase.STATES.PAID: 'pay'
            }
            transition_name = state_transition_map[self.state]

            bound_transition_method = getattr(self.related_document, transition_name)
            bound_transition_method()

    def clone_into_draft(self):
        copied_fields = {
            'customer': self.customer,
            'provider': self.provider,
            'currency': self.currency,
            'sales_tax_percent': self.sales_tax_percent,
            'sales_tax_name': self.sales_tax_name
        }

        clone = self.__class__._default_manager.create(**copied_fields)
        clone.state = self.STATES.DRAFT

        # clone entries too
        for entry in self._entries:
            entry_clone = entry.clone()
            document_type_name = self.__class__.__name__.lower()
            setattr(entry_clone, document_type_name, clone)
            entry_clone.save()

        clone.save()

        return clone

    def clean(self):
        super(BillingDocumentBase, self).clean()

        # The only change that is allowed if the document is in issued state
        # is the state chage from issued to paid
        # !! TODO: If _last_state == 'issued' and self.state == 'paid' || 'canceled'
        # it should also be checked that the other fields are the same bc.
        # right now a document can be in issued state and someone could
        # send a request which contains the state = 'paid' and also send
        # other changed fields and the request would be accepted bc. only
        # the state is verified.
        if self._last_state == self.STATES.ISSUED and\
           self.state not in [self.STATES.PAID, self.STATES.CANCELED]:
            msg = 'You cannot edit the document once it is in issued state.'
            raise ValidationError({NON_FIELD_ERRORS: msg})

        if self._last_state == self.STATES.CANCELED:
            msg = 'You cannot edit the document once it is in canceled state.'
            raise ValidationError({NON_FIELD_ERRORS: msg})

        # If it's in paid state => don't allow any changes
        if self._last_state == self.STATES.PAID:
            msg = 'You cannot edit the document once it is in paid state.'
            raise ValidationError(msg)

        if self.transactions.exclude(currency=self.transaction_currency).exists():
            message = 'There are unfinished transactions of this document that use a ' \
                      'different currency.'
            raise ValidationError({'transaction_currency': message})

    def clean_defaults(self):
        if not self.transaction_currency:
            self.transaction_currency = self.customer.currency or self.currency

        if not self.series:
            self.series = self.default_series

        # Generate the number
        if not self.number and self.state != BillingDocumentBase.STATES.DRAFT:
            self.number = self._generate_number()

        # Add tax info
        if not self.sales_tax_name:
            self.sales_tax_name = self.customer.sales_tax_name
        if not self.sales_tax_percent:
            self.sales_tax_percent = self.customer.sales_tax_percent

    def save(self, *args, **kwargs):
        # ToDo: Use AutoCleanModelMixin for this class and move clean_defaults() call
        # to full_clean
        self.clean_defaults()

        self._last_state = self.state

        with db_transaction.atomic():
            # Create pdf object
            if not self.pdf and self.state != self.STATES.DRAFT:
                self.pdf = PDF.objects.create(upload_path=self.get_pdf_upload_path(), dirty=1)

            super(BillingDocumentBase, self).save(*args, **kwargs)

    def _generate_number(self, default_starting_number=1):
        """Generates the number for a proforma/invoice."""
        default_starting_number = max(default_starting_number, 1)

        documents = self.__class__._default_manager.filter(
            provider=self.provider, series=self.series
        )
        if not documents.exists():
            # An invoice/proforma with this provider and series does not exist
            if self.series == self.default_series:
                return self._starting_number
            else:
                return default_starting_number
        else:
            # An invoice with this provider and series already exists
            max_existing_number = documents.aggregate(
                Max('number')
            )['number__max']
            if max_existing_number:
                if self._starting_number and self.series == self.default_series:
                    return max(max_existing_number + 1, self._starting_number)
                else:
                    return max_existing_number + 1
            else:
                return default_starting_number

    def series_number(self):
        if self.series:
            if self.number:
                return "%s-%d" % (self.series, self.number)
            else:
                return "%s-draft-id:%d" % (self.series, self.pk)

        else:
            return "draft-id:%d" % self.pk

    series_number.short_description = 'Number'
    series_number = property(series_number)

    def __str__(self):
        return u'%s %s => %s [%.2f %s]' % (self.series_number,
                                           self.provider.billing_name,
                                           self.customer.billing_name,
                                           self.total, self.currency)

    @property
    def updateable_fields(self):
        return ['customer', 'provider', 'due_date', 'issue_date', 'paid_date',
                'cancel_date', 'sales_tax_percent', 'sales_tax_name',
                'currency']

    @property
    def admin_change_url(self):
        url_base = 'admin:{app_label}_{klass}_change'.format(
            app_label=self._meta.app_label,
            klass=self.__class__.__name__.lower())
        url = reverse(url_base, args=(self.pk,))
        return mark_safe('<a href="{url}">{display_series}</a>'.format(
            url=url, display_series=self.series_number))

    @property
    def _entries(self):
        # entries iterator which replaces the invoice/proforma from the DB with
        # self. We need this in generate_pdf so that the data in PDF has the
        # lastest state for the document. Without this we get in template:
        #
        # invoice.issue_date != entry.invoice.issue_date
        #
        # which is obviously false.
        document_type_name = self.__class__.__name__  # Invoice or Proforma
        kwargs = {document_type_name.lower(): self}
        entries = DocumentEntry.objects.filter(**kwargs)
        for entry in entries:
            if document_type_name.lower() == 'invoice':
                entry.invoice = self
            if document_type_name.lower() == 'proforma':
                entry.proforma = self
            yield(entry)

    def get_template_context(self, state=None):
        customer = Customer(**self.archived_customer)
        provider = Provider(**self.archived_provider)
        if state is None:
            state = self.state

        return {
            'document': self,
            'provider': provider,
            'customer': customer,
            'entries': self._entries,
            'state': state
        }

    def get_template(self, state=None):
        provider_state_template = '{provider}/{kind}_{state}_pdf.html'.format(
            kind=self.kind, provider=self.provider.slug, state=state).lower()
        provider_template = '{provider}/{kind}_pdf.html'.format(
            kind=self.kind, provider=self.provider.slug).lower()
        generic_state_template = '{kind}_{state}_pdf.html'.format(
            kind=self.kind, state=state).lower()
        generic_template = '{kind}_pdf.html'.format(
            kind=self.kind).lower()
        _templates = [provider_state_template, provider_template,
                      generic_state_template, generic_template]

        templates = []
        for t in _templates:
            templates.append('billing_documents/' + t)

        return select_template(templates)

    def get_pdf_filename(self):
        return '{doc_type}_{series}-{number}.pdf'.format(
            doc_type=self.__class__.__name__,
            series=self.series,
            number=self.number
        )

    def get_pdf_upload_path(self):
        path_template = getattr(
            settings, 'SILVER_DOCUMENT_UPLOAD_PATH',
            'documents/{provider}/{doc.kind}/{issue_date}/{filename}'
        )

        context = {
            'doc': self,
            'filename': self.get_pdf_filename(),
            'provider': self.provider.slug,
            'customer': self.customer.slug,
            'issue_date': self.issue_date.strftime('%Y/%m/%d') if self.issue_date else "draft"
        }

        return path_template.format(**context)

    def generate_pdf(self, state=None, upload=True):
        # !!! ensure this is not called concurrently for the same document

        context = self.get_template_context(state)
        context['filename'] = self.get_pdf_filename()

        pdf_file_object = self.pdf.generate(template=self.get_template(state),
                                            context=context,
                                            upload=upload)

        return pdf_file_object

    def generate_html(self, state=None, request=None):
        context = self.get_template_context(state)
        template = self.get_template(state=context['state'])

        return template.render(context, request)

    def serialize_hook(self, hook):
        """
        Used to generate a skinny payload.
        """

        return {
            'hook': hook.dict(),
            'data': {
                'id': self.id
            }
        }

    @property
    def entries(self):
        raise NotImplementedError

    @property
    def total(self):
        if self._total is not None:
            return self._total

        return sum([entry.total for entry in self.entries])

    @property
    def total_before_tax(self):
        return sum([entry.total_before_tax for entry in self.entries])

    @property
    def tax_value(self):
        return sum([entry.tax_value for entry in self.entries])

    @property
    @require_transaction_currency_and_xe_rate
    def total_in_transaction_currency(self):
        if self._total_in_transaction_currency is not None:
            return self._total_in_transaction_currency

        return sum([entry.total_in_transaction_currency
                    for entry in self.entries])

    @property
    @require_transaction_currency_and_xe_rate
    def total_before_tax_in_transaction_currency(self):
        return sum([entry.total_before_tax_in_transaction_currency
                    for entry in self.entries])

    @property
    @require_transaction_currency_and_xe_rate
    def tax_value_in_transaction_currency(self):
        return sum([entry.tax_value_in_transaction_currency
                    for entry in self.entries])

    @property
    @require_transaction_currency_and_xe_rate
    def amount_paid_in_transaction_currency(self):
        Transaction = apps.get_model('silver.Transaction')

        return sum([transaction.amount
                    for transaction in self.transactions.filter(state=Transaction.States.Settled)])

    @property
    @require_transaction_currency_and_xe_rate
    def amount_pending_in_transaction_currency(self):
        Transaction = apps.get_model('silver.Transaction')

        return sum([transaction.amount
                    for transaction in self.transactions.filter(state=Transaction.States.Pending)])

    @property
    @require_transaction_currency_and_xe_rate
    def amount_to_be_charged_in_transaction_currency(self):
        Transaction = apps.get_model('silver.Transaction')

        return self.total_in_transaction_currency - sum([
            transaction.amount
            for transaction in self.transactions.filter(state__in=[
                Transaction.States.Initial,
                Transaction.States.Pending,
                Transaction.States.Settled
            ])
        ])


def create_transaction_for_document(document):
    # get a usable, recurring payment_method for the customer
    PaymentMethod = apps.get_model('silver.PaymentMethod')
    Transaction = apps.get_model('silver.Transaction')

    payment_methods = PaymentMethod.objects.filter(
        canceled=False,
        verified=True,
        customer=document.customer
    )
    for payment_method in payment_methods:
        try:
            return Transaction.objects.create(document=document,
                                              payment_method=payment_method)
        except ValidationError:
            continue


@receiver(post_transition)
def post_transition_callback(sender, instance, name, source, target, **kwargs):
    if not isinstance(instance, BillingDocumentBase):
        return

    document = instance
    setattr(document, '.recently_transitioned', target)

    document.save()


@receiver(post_save)
def post_document_save(sender, instance, created=False, **kwargs):
    if not isinstance(instance, BillingDocumentBase):
        return

    document = instance

    if not hasattr(document, '.recently_transitioned'):
        return

    # The document has been transitioned before being saved
    delattr(document, '.recently_transitioned')

    # Transition related document too, if needed
    document.sync_related_document_state()

    # Create a transaction if the document was recently issued
    if (document.state == BillingDocumentBase.STATES.ISSUED and
            settings.SILVER_AUTOMATICALLY_CREATE_TRANSACTIONS):
        # But only if there is no pending transaction
        Transaction = apps.get_model('silver', 'Transaction')

        # The related document might have the only reference to an existing transaction
        if not (document.related_document or document).transactions.filter(
            state__in=[Transaction.States.Pending,
                       Transaction.States.Initial,
                       Transaction.States.Settled]
        ):
            create_transaction_for_document(document)

    # Generate a PDF
    document.mark_for_generation()



# Copyright (c) 2017 Presslabs SRL
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import, unicode_literals

import uuid
import logging

from decimal import Decimal

from annoying.functions import get_object_or_None
from django.core.serializers.json import DjangoJSONEncoder
from django_fsm import FSMField, post_transition, transition

from django.core.exceptions import ValidationError
from django.core.validators import MinValueValidator
from django.db import models, transaction
from django.db.models import Q, JSONField
from django.db.models.signals import post_save
from django.dispatch import receiver
from django.utils import timezone
from django.utils.encoding import force_str
from django.utils.translation import gettext_lazy as _

from silver.models import Invoice, Proforma
from silver.models.transactions.codes import FAIL_CODES, REFUND_CODES, CANCEL_CODES
from silver.utils.international import currencies
from silver.utils.models import AutoDateTimeField, AutoCleanModelMixin

logger = logging.getLogger(__name__)


class Transaction(AutoCleanModelMixin,
                  models.Model):
    _provider = None

    amount = models.DecimalField(
        decimal_places=2, max_digits=12,
        validators=[MinValueValidator(Decimal('0.00'))]
    )
    currency = models.CharField(
        choices=currencies, max_length=4,
        help_text='The currency used for billing.'
    )

    class Meta:
        ordering = ['-id']

    class States:
        Initial = 'initial'
        Pending = 'pending'
        Settled = 'settled'
        Failed = 'failed'
        Canceled = 'canceled'
        Refunded = 'refunded'

        @classmethod
        def as_list(cls):
            return [getattr(cls, state) for state in vars(cls).keys() if
                    state[0].isupper()]

        @classmethod
        def as_choices(cls):
            return (
                (state, _(state.capitalize())) for state in cls.as_list()
            )

    external_reference = models.CharField(max_length=256, null=True, blank=True)
    data = JSONField(default=dict, null=True, blank=True, encoder=DjangoJSONEncoder)
    state = FSMField(max_length=8, choices=States.as_choices(),
                     default=States.Initial)

    proforma = models.ForeignKey("BillingDocumentBase", null=True, blank=True,
                                 on_delete=models.SET_NULL, related_name='proforma_transactions')
    invoice = models.ForeignKey("BillingDocumentBase", null=True, blank=True,
                                on_delete=models.SET_NULL, related_name='invoice_transactions')

    payment_method = models.ForeignKey('PaymentMethod', on_delete=models.PROTECT)
    uuid = models.UUIDField(default=uuid.uuid4)
    valid_until = models.DateTimeField(null=True, blank=True)
    last_access = models.DateTimeField(null=True, blank=True)

    created_at = models.DateTimeField(default=timezone.now)
    updated_at = AutoDateTimeField(default=timezone.now)

    fail_code = models.CharField(
        choices=[(code, code) for code in FAIL_CODES.keys()], max_length=64,
        null=True, blank=True
    )
    refund_code = models.CharField(
        choices=[(code, code) for code in REFUND_CODES.keys()], max_length=32,
        null=True, blank=True
    )
    cancel_code = models.CharField(
        choices=[(code, code) for code in CANCEL_CODES.keys()], max_length=32,

        null=True, blank=True
    )

    @property
    def final_fields(self):
        fields = ['proforma', 'invoice', 'uuid', 'payment_method', 'amount',
                  'currency', 'created_at']

        return fields

    def __init__(self, *args, **kwargs):
        self.form_class = kwargs.pop('form_class', None)

        super(Transaction, self).__init__(*args, **kwargs)

    @transition(field=state, source=States.Initial, target=States.Pending)
    def process(self):
        pass

    @transition(field=state, source=[States.Initial, States.Pending],
                target=States.Settled)
    def settle(self):
        pass

    @transition(field=state, source=[States.Initial, States.Pending],
                target=States.Canceled)
    def cancel(self, cancel_code='default', cancel_reason='Unknown cancel reason'):
        self.cancel_code = cancel_code
        logger.error(str(cancel_reason))

    @transition(field=state, source=[States.Initial, States.Pending],
                target=States.Failed)
    def fail(self, fail_code='default', fail_reason='Unknown fail reason'):
        self.fail_code = fail_code
        logger.error(str(fail_reason))

    @transition(field=state, source=States.Settled, target=States.Refunded)
    def refund(self, refund_code='default', refund_reason='Unknown refund reason'):
        self.refund_code = refund_code
        logger.error(str(refund_reason))

    @transaction.atomic()
    def save(self, *args, **kwargs):
        if not self.pk:
            # Creating a new Transaction so we lock the DB rows for related billing documents and
            # transactions
            if self.proforma:
                Proforma.objects.select_for_update().filter(pk=self.proforma.pk)
            elif self.invoice:
                Invoice.objects.select_for_update().filter(pk=self.invoice.pk)

            Transaction.objects.select_for_update().filter(Q(proforma=self.proforma) |
                                                           Q(invoice=self.invoice))

        super(Transaction, self).save(*args, **kwargs)

    def clean(self):
        # Validate documents
        document = self.document
        if not document:
            raise ValidationError(
                'The transaction must have at least one billing document '
                '(invoice or proforma).'
            )

        if document.state == document.STATES.DRAFT:
            raise ValidationError(
                'The transaction must have a non-draft billing document '
                '(invoice or proforma).'
            )

        if self.invoice and self.proforma:
            if self.invoice.related_document != self.proforma:
                raise ValidationError('Invoice and proforma are not related.')
        else:
            if self.invoice:
                self.proforma = self.invoice.related_document
            else:
                self.invoice = self.proforma.related_document

        if document.customer != self.customer:
            raise ValidationError(
                'Customer doesn\'t match with the one in documents.'
            )

        # New transaction
        if not self.pk:
            if document.state != document.STATES.ISSUED:
                raise ValidationError(
                    'Transactions can only be created for issued documents.'
                )

            if self.currency:
                if self.currency != self.document.transaction_currency:
                    message = "Transaction currency is different from it's document's "\
                              "transaction_currency."
                    raise ValidationError(message)
            else:
                self.currency = self.document.transaction_currency

            if (self.payment_method.allowed_currencies and
                    self.currency not in self.payment_method.allowed_currencies):
                message = 'Currency {} is not allowed by the payment method. Allowed currencies ' \
                          'are {}.'.format(
                              self.currency, self.payment_method.allowed_currencies
                          )
                raise ValidationError(message)
            if self.amount:
                if self.amount > self.document.amount_to_be_charged_in_transaction_currency:
                    message = "Amount is greater than the amount that should be charged in order " \
                              "to pay the billing document."
                    raise ValidationError(message)
            else:
                self.amount = self.document.amount_to_be_charged_in_transaction_currency
        else:
            # clean final fields
            errors = {}
            for field in self.final_fields:
                old_value = self.initial_state.get(field)
                current_value = self.current_state.get(field)

                if old_value is not None and old_value != current_value:
                    errors[field] = 'This field may not be modified.'
            if errors:
                raise ValidationError(errors)

    def full_clean(self, *args, **kwargs):
        # 'amount' and 'currency' are handled in our clean method
        kwargs['exclude'] = kwargs.get('exclude', []) + ['currency', 'amount']

        super(Transaction, self).full_clean(*args, **kwargs)

    @property
    def can_be_consumed(self):
        if self.valid_until and self.valid_until < timezone.now():
            return False

        if self.state != Transaction.States.Initial:
            return False

        return True

    @property
    def customer(self):
        return self.payment_method.customer

    @property
    def document(self):
        return self.invoice or self.proforma

    @document.setter
    def document(self, value):
        if isinstance(value, Invoice):
            self.invoice = value
        elif isinstance(value, Proforma):
            self.proforma = value
        else:
            raise ValueError(
                'The provided document is not an invoice or a proforma.'
            )

    @property
    def provider(self):
        return self._provider or self.document.provider

    @provider.setter
    def provider(self, provider):
        self._provider = provider

    @property
    def payment_processor(self):
        return self.payment_method.payment_processor

    def update_document_state(self):
        if (
            self.state == Transaction.States.Settled and
            not self.document.amount_to_be_charged_in_transaction_currency and
            self.document.state != self.document.STATES.PAID
        ):
            self.document.pay()

    def __str__(self):
        return force_str(self.uuid)


@receiver(post_transition)
def post_transition_callback(sender, instance, name, source, target, **kwargs):
    if issubclass(sender, Transaction):
        setattr(instance, '.recently_transitioned', target)


@receiver(post_save, sender=Transaction)
def post_transaction_save(sender, instance, **kwargs):
    transaction = instance

    if hasattr(transaction, '.recently_transitioned'):
        delattr(transaction, '.recently_transitioned')
        transaction.update_document_state()

    if hasattr(transaction, '.cleaned'):
        delattr(transaction, '.cleaned')

    if not getattr(transaction, 'previous_instance', None):
        # we know this instance is freshly made as it doesn't have an old_value
        logger.info('[Models][Transaction]: %s', {
            'detail': 'A transaction was created.',
            'transaction_id': transaction.id,
            'customer_id': transaction.customer.id,
            'invoice_id': transaction.invoice.id if transaction.invoice else None,
            'proforma_id':
                transaction.proforma.id if transaction.proforma else None
        })



# Copyright (c) 2017 Presslabs SRL
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

from django.forms import Form


class GenericTransactionForm(Form):
    def __init__(self, payment_method, transaction, request=None, *args, **kwargs):
        self.payment_method = payment_method
        self.transaction = transaction
        self.request = request

        super(GenericTransactionForm, self).__init__(*args, **kwargs)



import pytest

from silver.fixtures.factories import BillingLogFactory


@pytest.mark.django_db
def test_update_billing_log_when_creating_proforma_related_invoice():
    billing_log = BillingLogFactory.create(invoice=None)
    proforma = billing_log.proforma

    assert billing_log.invoice is None

    invoice = proforma.create_invoice()
    billing_log.refresh_from_db()
    assert billing_log.invoice == invoice



# Copyright (c) 2015 Presslabs SRL
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

from annoying.functions import get_object_or_None
from django_filters.rest_framework import DjangoFilterBackend

from rest_framework import generics, permissions, status
from rest_framework.generics import get_object_or_404
from rest_framework.response import Response

from silver.api.filters import PlanFilter
from silver.api.serializers.common import MeteredFeatureSerializer
from silver.api.serializers.plans_serializer import PlanSerializer
from silver.models import Plan, MeteredFeature


class PlanList(generics.ListCreateAPIView):

    permission_classes = (permissions.IsAuthenticated,)
    serializer_class = PlanSerializer
    queryset = Plan.objects.all().prefetch_related('metered_features')
    filter_backends = (DjangoFilterBackend,)
    filterset_class = PlanFilter


class PlanDetail(generics.RetrieveDestroyAPIView):
    permission_classes = (permissions.IsAuthenticated,)
    serializer_class = PlanSerializer
    model = Plan

    def get_object(self):
        pk = self.kwargs.get('pk', None)
        return get_object_or_404(Plan, pk=pk)

    def patch(self, request, *args, **kwargs):
        plan = get_object_or_404(Plan.objects, pk=self.kwargs.get('pk', None))
        name = request.data.get('name', None)
        generate_after = request.data.get('generate_after', None)
        plan.name = name or plan.name
        plan.generate_after = generate_after or plan.generate_after
        plan.save()
        return Response(PlanSerializer(plan, context={'request': request}).data,
                        status=status.HTTP_200_OK)

    def delete(self, request, *args, **kwargs):
        plan = get_object_or_404(Plan.objects, pk=self.kwargs.get('pk', None))
        plan.enabled = False
        plan.save()
        return Response({"deleted": not plan.enabled},
                        status=status.HTTP_200_OK)


class PlanMeteredFeatures(generics.ListAPIView):
    permission_classes = (permissions.IsAuthenticated,)
    serializer_class = MeteredFeatureSerializer
    model = MeteredFeature

    def get_queryset(self):
        plan = get_object_or_None(Plan, pk=self.kwargs['pk'])
        return plan.metered_features.all() if plan else None



# Copyright (c) 2015 Presslabs SRL
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

from datetime import timedelta
from dateutil.relativedelta import *


ONE_DAY = timedelta(days=1)
ONE_MONTH = relativedelta(months=1)


def next_month(date):
    return date + ONE_MONTH


def first_day_of_month(date):
    return date + relativedelta(day=1)


def last_day_of_month(date):
    return date + relativedelta(day=31)


def prev_month(date):
    return date - ONE_MONTH



# -*- coding: utf-8 -*-
# Generated by Django 1.11.20 on 2019-06-25 08:55
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('silver', '0051_auto_20190603_0755'),
    ]

    operations = [
        migrations.AddField(
            model_name='billingdocumentbase',
            name='is_storno',
            field=models.BooleanField(default=False),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.18 on 2019-10-28 12:54
from __future__ import unicode_literals

from django.db import migrations, models
import django_fsm


class Migration(migrations.Migration):

    dependencies = [
        ('silver', '0052_billingdocumentbase_is_storno'),
    ]

    operations = [
        migrations.AlterField(
            model_name='subscription',
            name='state',
            field=django_fsm.FSMField(choices=[('active', 'Active'), ('inactive', 'Inactive'), ('canceled', 'Canceled'), ('ended', 'Ended')], default='inactive', help_text='The state the subscription is in.', max_length=12),
        ),
        migrations.AlterField(
            model_name='transaction',
            name='fail_code',
            field=models.CharField(blank=True, choices=[(b'transaction_declined_by_bank', b'transaction_declined_by_bank'), (b'transaction_hard_declined', b'transaction_hard_declined'), (b'invalid_payment_method', b'invalid_payment_method'), (b'expired_payment_method', b'expired_payment_method'), (b'default', b'default'), (b'invalid_card', b'invalid_card'), (b'insufficient_funds', b'insufficient_funds'), (b'transaction_declined', b'transaction_declined'), (b'expired_card', b'expired_card'), (b'transaction_hard_declined_by_bank', b'transaction_hard_declined_by_bank'), (b'limit_exceeded', b'limit_exceeded')], max_length=32, null=True),
        ),
        migrations.AlterField(
            model_name='transaction',
            name='state',
            field=django_fsm.FSMField(choices=[('canceled', 'Canceled'), ('refunded', 'Refunded'), ('initial', 'Initial'), ('failed', 'Failed'), ('settled', 'Settled'), ('pending', 'Pending')], default='initial', max_length=8),
        ),
    ]



# Copyright (c) 2015 Presslabs SRL
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

from mock import MagicMock, patch, call

from django.core.management import call_command
from django.test import TestCase, override_settings

from silver.fixtures.factories import TransactionFactory, PaymentMethodFactory
from silver.fixtures.test_fixtures import (TriggeredProcessor, PAYMENT_PROCESSORS,
                                           triggered_processor)


@override_settings(PAYMENT_PROCESSORS=PAYMENT_PROCESSORS)
class TestExecuteTransactionsCommand(TestCase):
    def test_transaction_executing(self):
        payment_method = PaymentMethodFactory.create(
            payment_processor=triggered_processor,
            verified=True
        )

        transactions = TransactionFactory.create_batch(
            5, payment_method=payment_method
        )

        mock_execute = MagicMock()
        with patch.multiple(TriggeredProcessor,
                            execute_transaction=mock_execute):
            call_command('execute_transactions')

            for transaction in transactions:
                self.assertIn(call(transaction), mock_execute.call_args_list)

            self.assertEqual(mock_execute.call_count, len(transactions))

    def test_transaction_filtering(self):
        payment_method = PaymentMethodFactory.create(
            payment_processor=triggered_processor,
            verified=True
        )

        transactions = TransactionFactory.create_batch(
            5, payment_method=payment_method
        )

        filtered_transactions = [
            transactions[0], transactions[2], transactions[4]
        ]

        mock_execute = MagicMock()
        with patch.multiple(TriggeredProcessor,
                            execute_transaction=mock_execute):
            transactions_arg = [
                str(transaction.pk) for transaction in filtered_transactions
            ]
            call_command('execute_transactions',
                         '--transactions=%s' % ','.join(transactions_arg))

            for transaction in filtered_transactions:
                self.assertIn(call(transaction), mock_execute.call_args_list)

            self.assertEqual(mock_execute.call_count, len(filtered_transactions))

    @patch('silver.management.commands.execute_transactions.logger.error')
    def test_exception_logging(self, mock_logger):
        payment_method = PaymentMethodFactory.create(
            payment_processor=triggered_processor,
            verified=True
        )

        TransactionFactory.create(payment_method=payment_method)

        mock_execute = MagicMock()
        mock_execute.side_effect = Exception('This happened.')

        with patch.multiple(TriggeredProcessor,
                            execute_transaction=mock_execute):
            call_command('execute_transactions')
            expected_call = call(
                'Encountered exception while executing transaction with id=%s.',
                1, exc_info=True
            )

            self.assertEqual(expected_call, mock_logger.call_args)

    def test_skip_transaction_with_unverified_payment_method(self):
        payment_method = PaymentMethodFactory.create(
            payment_processor=triggered_processor,
            verified=False
        )

        TransactionFactory.create(payment_method=payment_method)

        mock_execute = MagicMock()
        with patch.multiple(TriggeredProcessor,
                            execute_transaction=mock_execute):
            call_command('execute_transactions')

            self.assertEqual(mock_execute.call_count, 0)



# Copyright (c) 2015 Presslabs SRL
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

from django.conf import settings
from django.utils.module_loading import import_string


class RateNotFound(Exception):
    def __init__(self, from_currency=None, to_currency=None, date=None):
        self.from_currency = from_currency
        self.to_currency = to_currency
        self.date = date

    def __str__(self):
        if not all([self.from_currency, self.to_currency]):
            return 'No rate was found.'

        if not self.date:
            return 'No rate for {} to {}.'.format(self.from_currency,
                                                  self.to_currency)

        return 'No rate for {} to {}, from {} was found.'.format(
            self.from_currency, self.to_currency, self.date
        )


class DummyConverter(object):
    def convert(self, amount, from_currency, to_currency, date):
        if from_currency != to_currency:
            raise RateNotFound(from_currency, to_currency, date)
        return amount

try:
    CurrencyConverter = import_string(settings.SILVER_CURRENCY_CONVERTER)()
except AttributeError:
    CurrencyConverter = DummyConverter()



# Copyright (c) 2015 Presslabs SRL
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

from datetime import date, timedelta
from decimal import Decimal

from six.moves import zip

from django.test import TestCase

from silver.models import DocumentEntry, Proforma, Invoice
from silver.fixtures.factories import (ProformaFactory, InvoiceFactory,
                                       DocumentEntryFactory, CustomerFactory)


class TestInvoice(TestCase):
    def test_pay_invoice_related_proforma_state_change_to_paid(self):
        proforma = ProformaFactory.create()
        proforma.issue()
        proforma.create_invoice()

        assert proforma.related_document.state == Invoice.STATES.ISSUED

        proforma.related_document.pay()

        assert proforma.related_document.state == Invoice.STATES.PAID
        assert proforma.state == Proforma.STATES.PAID

    def test_clone_invoice_into_draft(self):
        invoice = InvoiceFactory.create()
        invoice.issue()
        invoice.pay()

        entries = DocumentEntryFactory.create_batch(3)
        invoice.invoice_entries.add(*entries)

        clone = invoice.clone_into_draft()

        assert clone.state == Invoice.STATES.DRAFT
        assert clone.paid_date is None
        assert clone.issue_date is None
        assert clone.related_document is None
        assert (clone.series != invoice.series or
                clone.number != invoice.number)
        assert clone.sales_tax_percent == invoice.sales_tax_percent
        assert clone.sales_tax_name == invoice.sales_tax_name

        assert not clone.archived_customer
        assert not clone.archived_provider
        assert clone.customer == invoice.customer
        assert clone.provider == invoice.provider

        assert clone.currency == invoice.currency
        assert clone._last_state == clone.state
        assert clone.pk != invoice.pk
        assert clone.id != invoice.id
        assert not clone.pdf

        assert clone.invoice_entries.count() == 3
        assert invoice.invoice_entries.count() == 3

        entry_fields = [entry.name for entry in DocumentEntry._meta.get_fields()]
        for clone_entry, original_entry in zip(clone.invoice_entries.all(),
                                               invoice.invoice_entries.all()):
            for entry in entry_fields:
                if entry not in ('id', 'proforma', 'invoice'):
                    assert getattr(clone_entry, entry) == \
                        getattr(original_entry, entry)
        assert invoice.state == Invoice.STATES.PAID

    def test_cancel_issued_invoice_with_related_proforma(self):
        proforma = ProformaFactory.create()
        proforma.issue()

        if not proforma.related_document:
            proforma.create_invoice()

        proforma.related_document.cancel()

        assert proforma.related_document.state == proforma.state == Invoice.STATES.CANCELED

    def _get_decimal_places(self, number):
        return max(0, -number.as_tuple().exponent)

    def test_invoice_total_decimal_points(self):
        invoice_entries = DocumentEntryFactory.create_batch(3)
        invoice = InvoiceFactory.create(invoice_entries=invoice_entries)

        assert self._get_decimal_places(invoice.total) == 2

    def test_invoice_total_before_tax_decimal_places(self):
        invoice_entries = DocumentEntryFactory.create_batch(3)
        invoice = InvoiceFactory.create(invoice_entries=invoice_entries)

        invoice.sales_tax_percent = Decimal('20.00')

        assert self._get_decimal_places(invoice.total_before_tax) == 2

    def test_invoice_tax_value_decimal_places(self):
        invoice_entries = DocumentEntryFactory.create_batch(3)
        invoice = InvoiceFactory.create(invoice_entries=invoice_entries)

        invoice.sales_tax_percent = Decimal('20.00')

        assert self._get_decimal_places(invoice.tax_value) == 2

    def test_invoice_total_with_tax_integrity(self):
        invoice_entries = DocumentEntryFactory.create_batch(5)
        invoice = InvoiceFactory.create(invoice_entries=invoice_entries)

        invoice.sales_tax_percent = Decimal('20.00')

        self.assertEqual(invoice.total, invoice.total_before_tax + invoice.tax_value)

    def test_draft_invoice_series_number(self):
        invoice = InvoiceFactory.create()
        invoice.number = None

        assert invoice.series_number == '%s-draft-id:%d' % (invoice.series,
                                                            invoice.pk)

        invoice.series = None

        assert invoice.series_number == 'draft-id:%d' % invoice.pk

    def test_issues_invoice_series_number(self):
        invoice = InvoiceFactory.create(state=Invoice.STATES.ISSUED)

        assert invoice.series_number == '%s-%s' % (invoice.series,
                                                   invoice.number)

    def test_invoice_due_today_queryset(self):
        invoices = InvoiceFactory.create_batch(5)

        invoices[0].due_date = date.today()
        invoices[0].save()

        invoices[1].due_date = date.today()
        invoices[1].issue()

        invoices[2].due_date = date.today() - timedelta(days=1)
        invoices[2].issue()
        invoices[2].pay()

        invoices[3].due_date = date.today()
        invoices[3].issue()
        invoices[3].cancel()

        invoices[4].due_date = date.today() + timedelta(days=1)
        invoices[4].issue()

        queryset = Invoice.objects.due_today()

        assert queryset.count() == 1
        assert invoices[1] in queryset

    def test_invoice_due_this_month_queryset(self):
        invoices = InvoiceFactory.create_batch(4)

        invoices[0].due_date = date.today().replace(day=20)
        invoices[0].issue()

        invoices[1].due_date = date.today().replace(day=1)
        invoices[1].issue()

        invoices[2].due_date = date.today() - timedelta(days=31)
        invoices[2].issue()

        invoices[3].issue()
        invoices[3].cancel()

        queryset = Invoice.objects.due_this_month()

        assert queryset.count() == 2
        for invoice in invoices[:2]:
            assert invoice in queryset

    def test_invoice_overdue_queryset(self):
        invoices = InvoiceFactory.create_batch(3)

        invoices[0].due_date = date.today() - timedelta(days=1)
        invoices[0].issue()

        invoices[1].due_date = date.today() - timedelta(days=3)
        invoices[1].issue()

        invoices[2].due_date = date.today() - timedelta(days=31)
        invoices[2].issue()
        invoices[2].pay()

        queryset = Invoice.objects.overdue()

        assert queryset.count() == 2
        for invoice in invoices[:2]:
            assert invoice in queryset

    def test_invoice_overdue_since_last_month_queryset(self):
        invoices = InvoiceFactory.create_batch(3)

        invoices[0].due_date = date.today().replace(day=1)
        invoices[0].issue()

        invoices[1].due_date = date.today() - timedelta(days=31)
        invoices[1].issue()

        queryset = Invoice.objects.overdue_since_last_month()

        assert queryset.count() == 1
        assert invoices[1] in queryset

    def test_customer_currency_used_for_transaction_currency(self):
        customer = CustomerFactory.create(currency='EUR')
        invoice = InvoiceFactory.create(customer=customer,
                                        transaction_currency=None)

        self.assertEqual(invoice.transaction_currency, 'EUR')

    def test_invoice_currency_used_for_transaction_currency(self):
        customer = CustomerFactory.create(currency=None)
        invoice = InvoiceFactory.create(customer=customer,
                                        currency='EUR',
                                        transaction_currency=None)

        self.assertEqual(invoice.transaction_currency, 'EUR')

    def test_invoice_create_storno_from_canceled_state(self):
        invoice = InvoiceFactory.create(invoice_entries=[DocumentEntryFactory.create()])
        invoice.issue()
        invoice.cancel()
        storno = invoice.create_storno()

        assert -invoice.total == storno.total != 0
        assert storno.related_document == invoice
        assert invoice.customer == storno.customer
        assert invoice.provider == storno.provider

    def test_invoice_create_storno_from_paid_state(self):
        invoice = InvoiceFactory.create(invoice_entries=[DocumentEntryFactory.create()])
        invoice.issue()
        invoice.pay()
        storno = invoice.create_storno()

        assert -invoice.total == storno.total != 0
        assert storno.related_document == invoice
        assert invoice.customer == storno.customer
        assert invoice.provider == storno.provider

    def test_invoice_create_storno_from_unallowed_states(self):
        invoice = InvoiceFactory.create()
        assert invoice.state == invoice.STATES.DRAFT
        self.assertRaises(ValueError, invoice.create_storno)

        invoice.issue()
        self.assertRaises(ValueError, invoice.create_storno)

    def test_invoice_storno_issue(self):
        invoice = InvoiceFactory.create()
        invoice.issue()
        invoice.pay()
        storno = invoice.create_storno()

        storno.issue()
        assert storno.state == storno.STATES.ISSUED
        assert storno.issue_date == date.today()
        assert not storno.due_date



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import uuid
from itertools import chain

from django.db import migrations, models
import django.db.models.deletion

import silver.models.documents.pdf


class Migration(migrations.Migration):

    dependencies = [
        ('silver', '0035_auto_20170206_0941'),
    ]

    def move_pdf_from_documents_to_model(apps, schema_editor):
        db_alias = schema_editor.connection.alias

        Invoice = apps.get_model('silver', 'Invoice')
        Proforma = apps.get_model('silver', 'Proforma')
        PDF = apps.get_model('silver', 'PDF')

        for document in chain(
            Invoice.objects.using(db_alias).exclude(state='draft'),
            Proforma.objects.using(db_alias).exclude(state='draft'),
        ):
            pdf_object = PDF.objects.using(db_alias).create()
            pdf_object.pdf_file = document.pdf_old
            pdf_object.save(using=db_alias)

            document.pdf = pdf_object
            document.save(using=db_alias)

    def move_pdf_from_model_to_documents(apps, schema_editor):
        db_alias = schema_editor.connection.alias

        Invoice = apps.get_model('silver', 'Invoice')
        Proforma = apps.get_model('silver', 'Proforma')

        for document in chain(
            Invoice.objects.using(db_alias).exclude(state='draft'),
            Proforma.objects.using(db_alias).exclude(state='draft'),
        ):
            document.pdf_old = document.pdf.pdf_file
            document.save()

    operations = [
        migrations.CreateModel(
            name='PDF',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('uuid', models.UUIDField(default=uuid.uuid4, unique=True)),
                ('pdf_file', models.FileField(upload_to=silver.models.documents.pdf.get_upload_path, null=True, editable=False, blank=True)),
                ('dirty', models.BooleanField(default=False)),
                ('upload_path', models.TextField(null=True, blank=True)),
            ],
        ),
        migrations.RenameField(
            model_name='invoice',
            old_name='pdf',
            new_name='pdf_old'
        ),
        migrations.RenameField(
            model_name='proforma',
            old_name='pdf',
            new_name='pdf_old'
        ),

        migrations.AddField(
            model_name='invoice',
            name='pdf',
            field=models.ForeignKey(to='silver.PDF', null=True, on_delete=django.db.models.deletion.SET_NULL),
        ),
        migrations.AddField(
            model_name='proforma',
            name='pdf',
            field=models.ForeignKey(to='silver.PDF', null=True, on_delete=django.db.models.deletion.SET_NULL),
        ),

        migrations.RunPython(move_pdf_from_documents_to_model,
                             move_pdf_from_model_to_documents),

        migrations.RemoveField(
            model_name='invoice',
            name='pdf_old',
        ),
        migrations.RemoveField(
            model_name='proforma',
            name='pdf_old',
        ),
    ]



# Copyright (c) 2015 Presslabs SRL
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

import json
import pytest

from rest_framework import status
from rest_framework.reverse import reverse
from rest_framework.test import APITestCase

from django.conf import settings

from silver.models import Customer
from silver.fixtures.factories import AdminUserFactory, CustomerFactory
from silver.tests.utils import build_absolute_test_url


class TestCustomerEndpoints(APITestCase):
    def setUp(self):
        admin_user = AdminUserFactory.create()
        self.client.force_authenticate(user=admin_user)
        self.complete_data = {
            "customer_reference": "123456",
            "first_name": "Bruce",
            "last_name": "Wayne",
            "company": "Wayne Enterprises",
            "email": "bruce@wayneenterprises.com",
            "address_1": "Batcave St.",
            "address_2": "Some other address info",
            "city": "Gotham",
            "state": "SomeState",
            "zip_code": "1111",
            "country": "US",
            "phone": "+40000000000",
            "currency": "USD",
            "extra": "What is there more to say?",
            "sales_tax_number": "RO5555555",
            "sales_tax_name": "VAT",
            "sales_tax_percent": '3.00',
            "payment_due_days": 5,
            "consolidated_billing": False,
            "meta": {'water': ['plants', '5']},
            "payment_methods": build_absolute_test_url(reverse('payment-method-list',
                                                               kwargs={'customer_pk': 1})),
            "transactions": build_absolute_test_url(reverse('transaction-list',
                                                            kwargs={'customer_pk': 1}))
        }

    def test_create_post_customer(self):
        url = reverse('customer-list')

        response = self.client.post(url, json.dumps(self.complete_data),
                                    content_type='application/json')
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

    def test_create_post_customer_without_required_field(self):
        url = reverse('customer-list')

        required_fields = ['first_name', 'last_name', 'address_1', 'city',
                           'country']

        for field in required_fields:
            temp_data = self.complete_data.copy()
            try:
                temp_data.pop(field)
            except KeyError:
                pytest.xfail('Customer required field %s not provided in the'
                             'complete test data.' % field)

            response = self.client.post(url, json.dumps(temp_data),
                                        content_type='application/json')

            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
            assert (response.data == {field: ['This field may not be blank.']} or
                    response.data == {field: ['This field is required.']})

    def test_get_customer_list(self):
        CustomerFactory.create_batch(settings.API_PAGE_SIZE * 2)

        url = reverse('customer-list')

        response = self.client.get(url)

        full_url = None
        for field in response.data:
            full_url = field.get('url', None)
            if full_url:
                break
        if full_url:
            domain = full_url.split('/')[2]
            full_url = full_url.split(domain)[0] + domain + url

        assert response.status_code == status.HTTP_200_OK
        assert response['link'] == \
            ('<' + full_url + '?page=2>; rel="next", ' +
             '<' + full_url + '?page=1>; rel="first", ' +
             '<' + full_url + '?page=2> rel="last"')

        response = self.client.get(url + '?page=2')

        assert response.status_code == status.HTTP_200_OK
        assert response['link'] == \
            ('<' + full_url + '>; rel="prev", ' +
             '<' + full_url + '?page=1>; rel="first", ' +
             '<' + full_url + '?page=2> rel="last"')

    def test_get_customer_detail(self):
        customer = CustomerFactory.create()

        url = reverse('customer-detail',
                      kwargs={'customer_pk': customer.pk})

        response = self.client.get(url)

        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertNotEqual(response.data, [])
        self.assertEqual(response.data['phone'], customer.phone)

    def test_get_customer_detail_unexisting(self):
        url = reverse('customer-detail',
                      kwargs={'customer_pk': 42})

        response = self.client.get(url)

        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)
        self.assertEqual(response.data, {u'detail': u'Not found.'})

    def test_delete_customer(self):
        customer = CustomerFactory.create()

        url = reverse('customer-detail', kwargs={'customer_pk': customer.pk})
        response = self.client.delete(url)

        assert response.status_code == status.HTTP_204_NO_CONTENT
        assert Customer.objects.all().count() == 0

    def test_delete_unexisting_customer(self):
        url = reverse('customer-detail', kwargs={'customer_pk': 42})
        response = self.client.delete(url)

        assert response.status_code == status.HTTP_404_NOT_FOUND

    def test_edit_put_customer(self):
        customer = CustomerFactory.create()

        changed_data = self.complete_data.copy()

        unchanged_fields = ['email', 'address_2']
        ignore_fields = ['url', 'id', 'subscriptions', 'payment_methods',
                         'transactions']
        for field in unchanged_fields:
            changed_data.pop(field)

        url = reverse('customer-detail', kwargs={'customer_pk': customer.pk})

        response = self.client.put(url, data=json.dumps(changed_data),
                                   content_type='application/json')

        self.assertEqual(response.status_code, status.HTTP_200_OK)

        for e in ignore_fields:
            response.data.pop(e)
        for field in response.data:
            if field not in unchanged_fields:
                self.assertEqual(response.data[field],
                                 self.complete_data[field])

    def test_edit_patch_customer(self):
        customer = CustomerFactory.create()

        changed_data = self.complete_data.copy()
        unchanged_fields = ['email', 'zip_code', 'company', 'phone',
                            'payment_due_days']
        ignore_fields = ['url', 'id', 'subscriptions', 'payment_methods',
                         'transactions']
        for field in unchanged_fields:
            changed_data.pop(field)

        url = reverse('customer-detail', kwargs={'customer_pk': customer.pk})

        response = self.client.patch(url, data=json.dumps(changed_data),
                                     content_type='application/json')

        self.assertEqual(response.status_code, status.HTTP_200_OK)

        for e in ignore_fields:
            response.data.pop(e)
        for field in response.data:
            if field not in unchanged_fields:
                self.assertEqual(response.data[field],
                                 self.complete_data[field])



# -*- coding: utf-8 -*-
# Generated by Django 1.11.3 on 2017-08-11 11:01
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('silver', '0038_auto_20170724_0708'),
    ]

    operations = [
        migrations.AddField(
            model_name='billinglog',
            name='total',
            field=models.DecimalField(blank=True, decimal_places=2, max_digits=12, null=True),
        ),
    ]



# Copyright (c) 2017 Presslabs SRL
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

from django.conf import settings
from django.template.loader import select_template
from django.utils.deconstruct import deconstructible
from django.utils.encoding import force_str
from django.utils.module_loading import import_string
from django.utils.text import slugify


def get_instance(name):
    data = settings.PAYMENT_PROCESSORS[name]
    klass = import_string(data['class'])
    kwargs = data.get('setup_data', {})
    return klass(name, **kwargs)


def get_all_instances():
    choices = []
    for processor_import_path in settings.PAYMENT_PROCESSORS.keys():
        choices.append(get_instance(processor_import_path))
    return choices


@deconstructible
class PaymentProcessorBase(object):
    form_class = None
    template_slug = None
    payment_method_class = None
    transaction_view_class = None
    allowed_currencies = ()

    def __init__(self, name):
        self.name = name

    def get_view(self, transaction, request, **kwargs):
        assert self.transaction_view_class, 'You must specify a `transaction_view_class` ' \
                                            'attribute for the {} class.'.format(
                                                self.__class__.__name__
                                            )

        kwargs.update({
            'form': self.get_form(transaction, request),
            'template': self.get_template(transaction),
            'transaction': transaction,
            'request': request
        })
        return self.transaction_view_class.as_view(**kwargs)

    def get_form(self, transaction, request):
        form = None
        if self.form_class:
            form = self.form_class(
                payment_method=transaction.payment_method,
                transaction=transaction, request=request
            )
        return form

    def get_template(self, transaction):
        provider = transaction.document.provider
        provider_slug = slugify(provider.company or provider.name)

        template = select_template([
            'forms/{}/{}_transaction_form.html'.format(
                self.template_slug,
                provider_slug
            ),
            'forms/{}/transaction_form.html'.format(
                self.template_slug
            ),
            'forms/transaction_form.html'
        ])

        return template

    def handle_transaction_response(self, transaction, request):
        """
            :param transaction: A Silver Transaction object.
            :param request: A Django request object. It should contain POST (or GET) data about the
            transaction, which will be used update the Silver Transaction.

            This method should update the transaction info (external reference, state ...) after the
            first HTTP response from the payment gateway.

            It will automatically be called by the `complete_payment_view`.

            If not needed, one can `pass` it or just `return`.
        """

        raise NotImplementedError

    def __repr__(self):
        return self.name

    def __str__(self):
        return force_str(self.name)

    def __eq__(self, other):
        return self.__class__ is other.__class__

    def __ne__(self, other):
        return not self.__eq__(other)



# Copyright (c) 2016 Presslabs SRL
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

from pyvat import is_vat_number_format_valid

from django.conf import settings
from django.core.exceptions import ValidationError
from django.core.validators import MinValueValidator
from django.db import models
from django.utils.html import escape
from django.utils.safestring import mark_safe
from django.urls import reverse

from silver.utils.international import currencies
from silver.models.billing_entities.base import BaseBillingEntity
from silver.validators import validate_reference


PAYMENT_DUE_DAYS = getattr(settings, 'SILVER_DEFAULT_DUE_DAYS', 5)


class Customer(BaseBillingEntity):
    class Meta:
        index_together = (('first_name', 'last_name', 'company'),)
        ordering = ['first_name', 'last_name', 'company']

    first_name = models.CharField(
        max_length=128,
        help_text='The customer\'s first name.'
    )
    last_name = models.CharField(
        max_length=128,
        help_text='The customer\'s last name.'
    )

    payment_due_days = models.PositiveIntegerField(
        default=PAYMENT_DUE_DAYS,
        help_text='Due days for generated proforma/invoice.'
    )
    consolidated_billing = models.BooleanField(
        default=False, help_text='A flag indicating consolidated billing.'
    )
    customer_reference = models.CharField(
        max_length=256, blank=True, null=True, validators=[validate_reference],
        help_text="It's a reference to be passed between silver and clients. "
                  "It usually points to an account ID."
    )
    sales_tax_number = models.CharField(max_length=64, blank=True, null=True)
    sales_tax_percent = models.DecimalField(
        max_digits=4, decimal_places=2, null=True, blank=True,
        validators=[MinValueValidator(0.0)],
        help_text="Whenever to add sales tax. "
                  "If null, it won't show up on the invoice."
    )
    sales_tax_name = models.CharField(
        max_length=64, null=True, blank=True,
        help_text="Sales tax name (eg. 'sales tax' or 'VAT')."
    )

    currency = models.CharField(
        choices=currencies, max_length=4, null=True, blank=True,
        help_text="Used to enforce a certain currency when making transactions"
                  "for the customer."
    )

    def __init__(self, *args, **kwargs):
        archived_name = None
        if 'name' in kwargs:
            archived_name = kwargs.pop('name')

        super(Customer, self).__init__(*args, **kwargs)

        if archived_name:
            self.first_name, self.last_name = archived_name.split()

        company_field = self._meta.get_field("company")
        company_field.help_text = "The company to which the bill is issued."

    def clean(self):
        if (self.sales_tax_number and
            is_vat_number_format_valid(self.sales_tax_number,
                                       self.country) is False):
            raise ValidationError(
                {'sales_tax_number': 'The sales tax number is not valid.'}
            )

    def get_archivable_field_values(self):
        base_fields = super(Customer, self).get_archivable_field_values()
        customer_fields = ['first_name', 'last_name', 'customer_reference', 'consolidated_billing',
                           'payment_due_days', 'sales_tax_number', 'sales_tax_percent']
        fields_dict = {field: getattr(self, field, '') for field in
                       customer_fields}
        base_fields.update(fields_dict)
        return base_fields

    @property
    def name(self):
        return u"%s %s" % (self.first_name, self.last_name)

    @property
    def admin_change_url(self):
        display = escape(self.name)
        if self.company:
            display += "<hr> " + escape(self.company)

        link = reverse("admin:silver_customer_change", args=[self.pk])
        return mark_safe('<a href="%s">%s</a>' % (link, display))

    def __str__(self):
        return self.name



# Copyright (c) 2015 Presslabs SRL
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

import logging
import argparse

from datetime import datetime as dt

from django.core.management.base import BaseCommand
from django.utils import translation

from silver.documents_generator import DocumentsGenerator
from silver.models import Subscription


logger = logging.getLogger(__name__)


def date(date_str):
    try:
        return dt.strptime(date_str, "%Y-%m-%d").date()
    except ValueError:
        msg = "Not a valid date: '{date_str}'. "\
              "Expected format: YYYY-MM-DD.".format(date_str=date_str)
        raise argparse.ArgumentTypeError(msg)


class Command(BaseCommand):
    help = 'Generates the billing documents (Invoices, Proformas).'

    def add_arguments(self, parser):
        parser.add_argument('--subscription',
                            action='store', dest='subscription_id', type=int,
                            help='The id of ths subscription to be billed.')
        parser.add_argument('--date',
                            action='store', dest='billing_date', type=date,
                            help='The billing date (format YYYY-MM-DD).')

    def handle(self, *args, **options):
        translation.activate('en-us')

        billing_date = options['billing_date']

        docs_generator = DocumentsGenerator()
        if options['subscription_id']:
            try:
                subscription_id = options['subscription_id']
                logger.info('Generating for subscription with id=%s; '
                            'billing_date=%s.', subscription_id,
                            billing_date)

                subscription = Subscription.objects.get(id=subscription_id)
                docs_generator.generate(subscription=subscription,
                                        billing_date=billing_date)
                self.stdout.write('Done. You can have a Club-Mate now. :)')
            except Subscription.DoesNotExist:
                msg = 'The subscription with the provided id does not exist.'
                self.stdout.write(msg)
        else:
            logger.info('Generating for all the available subscriptions; '
                        'billing_date=%s.', billing_date)

            docs_generator.generate(billing_date=billing_date)
            self.stdout.write('Done. You can have a Club-Mate now. :)')




def spec_archived_customer(customer):
    return {
        'company': customer.company,
        'email': customer.email,
        'address_1': customer.address_1,
        'address_2': customer.address_2,
        'city': customer.city,
        'country': customer.country,
        'state': customer.state,
        'zip_code': customer.zip_code,
        'extra': customer.extra,
        'meta': customer.meta,
        'first_name': customer.first_name,
        'last_name': customer.last_name,
        'customer_reference': customer.customer_reference,
        'consolidated_billing': bool(customer.consolidated_billing),
        'payment_due_days': int(customer.payment_due_days),
        'sales_tax_number': customer.sales_tax_number,
        'sales_tax_percent': "%.2f" % customer.sales_tax_percent
    }



# Copyright (c) 2015 Presslabs SRL
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

from itertools import chain

from celery import group, shared_task
from celery_once import QueueOnce
from redis.exceptions import LockError

from django.conf import settings
from django.utils import timezone

from silver.documents_generator import DocumentsGenerator
from silver.models import Invoice, Proforma, Transaction, BillingDocumentBase
from silver.payment_processors.mixins import PaymentProcessorTypes
from silver.vendors.redis_server import redis


PDF_GENERATION_TIME_LIMIT = getattr(settings, 'PDF_GENERATION_TIME_LIMIT',
                                    60)  # default 60s


@shared_task(base=QueueOnce, once={'graceful': True},
             time_limit=PDF_GENERATION_TIME_LIMIT)
def generate_pdf(document_id, document_type):
    document = BillingDocumentBase.objects.get(id=document_id, kind=document_type)

    document.generate_pdf()


@shared_task(ignore_result=True)
def generate_pdfs():
    dirty_documents = chain(Invoice.objects.filter(pdf__dirty__gt=0),
                            Proforma.objects.filter(pdf__dirty__gt=0))

    # Generate PDFs in parallel
    group(generate_pdf.s(document.id, document.kind)
          for document in dirty_documents)()


DOCS_GENERATION_TIME_LIMIT = getattr(settings, 'DOCS_GENERATION_TIME_LIMIT',
                                     60 * 60)  # default 60m


@shared_task(base=QueueOnce, once={'graceful': True},
             time_limit=DOCS_GENERATION_TIME_LIMIT, ignore_result=True)
def generate_billing_documents(billing_date=None):
    if not billing_date:
        billing_date = timezone.now().date()

    DocumentsGenerator().generate(billing_date=billing_date)


FETCH_TRANSACTION_STATUS_TIME_LIMIT = getattr(settings, 'FETCH_TRANSACTION_STATUS_TIME_LIMIT',
                                              60)  # default 60s


@shared_task(base=QueueOnce, once={'graceful': True},
             time_limit=FETCH_TRANSACTION_STATUS_TIME_LIMIT)
def fetch_transaction_status(transaction_id):
    transaction = Transaction.objects.filter(pk=transaction_id).first()
    if not transaction:
        return

    payment_processor = transaction.payment_method.get_payment_processor()
    if payment_processor.type != PaymentProcessorTypes.Triggered:
        return

    payment_processor.fetch_transaction_status(transaction)


@shared_task(ignore_result=True)
def fetch_transactions_status(transaction_ids=None):
    eligible_transactions = Transaction.objects.filter(state=Transaction.States.Pending)

    if transaction_ids:
        eligible_transactions = eligible_transactions.filter(pk__in=transaction_ids)

    group(fetch_transaction_status.s(transaction.id) for transaction in eligible_transactions)()


EXECUTE_TRANSACTION_TIME_LIMIT = getattr(settings, 'EXECUTE_TRANSACTION_TIME_LIMIT',
                                         60)  # default 60s


@shared_task(base=QueueOnce, once={'graceful': True},
             time_limit=EXECUTE_TRANSACTION_TIME_LIMIT)
def execute_transaction(transaction_id):
    transaction = Transaction.objects.filter(pk=transaction_id).first()
    if not transaction:
        return

    if not transaction.payment_method.verified or transaction.payment_method.canceled:
        return

    payment_processor = transaction.payment_method.get_payment_processor()
    if payment_processor.type != PaymentProcessorTypes.Triggered:
        return

    payment_processor.process_transaction(transaction)


@shared_task(ignore_result=True)
def execute_transactions(transaction_ids=None):
    executable_transactions = Transaction.objects.filter(state=Transaction.States.Initial)

    if transaction_ids:
        executable_transactions = executable_transactions.filter(pk__in=transaction_ids)

    group(execute_transaction.s(transaction.id) for transaction in executable_transactions)()



# Copyright (c) 2017 Pressinfra SRL
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

import logging

from itertools import chain

from django.core.management.base import BaseCommand

from silver.models import Invoice, Proforma


logger = logging.getLogger(__name__)


class Command(BaseCommand):
    help = 'Generates the billing documents (Invoices, Proformas).'

    def handle(self, *args, **options):
        for document in chain(Invoice.objects.filter(pdf__dirty__gt=0),
                              Proforma.objects.filter(pdf__dirty__gt=0)):
            try:
                document.generate_pdf()
            except:
                logger.exception('Encountered exception while generating PDF for document '
                                 'with id=%s.', document.id)



# Copyright (c) 2017 Presslabs SRL
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import, unicode_literals

from typing import Union

from itertools import chain

from annoying.functions import get_object_or_None
from cryptography.fernet import InvalidToken, Fernet
from django.core.serializers.json import DjangoJSONEncoder
from django_fsm import TransitionNotAllowed
from model_utils.managers import InheritanceManager

from django.conf import settings
from django.core.exceptions import ValidationError
from django.db import models
from django.db.models.signals import post_save, pre_save
from django.dispatch import receiver
from django.utils import timezone

from silver import payment_processors
from silver.models import Invoice, Proforma
from silver.models.billing_entities import Customer
from silver.models.transactions import Transaction


class PaymentMethodInvalid(Exception):
    pass


class PaymentMethod(models.Model):
    class PaymentProcessors:
        @classmethod
        def as_choices(cls):
            for name in settings.PAYMENT_PROCESSORS.keys():
                yield (name, name)

        @classmethod
        def as_list(cls):
            return [name for name in settings.PAYMENT_PROCESSORS.keys()]

    payment_processor = models.CharField(choices=PaymentProcessors.as_choices(),
                                         blank=False, null=False, max_length=256)
    customer = models.ForeignKey(Customer, models.CASCADE)
    added_at = models.DateTimeField(default=timezone.now)
    data = models.JSONField(blank=True, null=True, default=dict, encoder=DjangoJSONEncoder)

    verified = models.BooleanField(default=False)
    canceled = models.BooleanField(default=False)

    valid_until = models.DateTimeField(null=True, blank=True)
    display_info = models.CharField(max_length=256, null=True, blank=True)

    objects = InheritanceManager()

    class Meta:
        ordering = ['-id']

    @property
    def final_fields(self):
        return ['payment_processor', 'customer', 'added_at']

    @property
    def irreversible_fields(self):
        return ['verified', 'canceled']

    def __init__(self, *args, **kwargs):
        super(PaymentMethod, self).__init__(*args, **kwargs)

        if self.id:
            try:
                payment_method_class = self.get_payment_processor().payment_method_class

                if payment_method_class:
                    self.__class__ = payment_method_class
            except AttributeError:
                pass

    @property
    def transactions(self):
        return self.transaction_set.all()

    def get_payment_processor(self):
        return payment_processors.get_instance(self.payment_processor)

    def delete(self, using=None):
        if not self.state == self.States.Uninitialized:
            self.remove()

        super(PaymentMethod, self).delete(using=using)

    def encrypt_data(self, data: Union[str, bytes]) -> str:
        if isinstance(data, str):
            data = data.encode(encoding="utf-8")

        key = settings.PAYMENT_METHOD_SECRET
        return Fernet(key).encrypt(data).decode('utf-8')

    def decrypt_data(self, crypted_data: Union[str, bytes]) -> str:
        if not crypted_data:
            return ""

        if isinstance(crypted_data, str):
            crypted_data = crypted_data.encode(encoding="utf-8")

        key = settings.PAYMENT_METHOD_SECRET

        return Fernet(key).decrypt(crypted_data).decode("utf-8")

    def cancel(self):
        if self.canceled:
            raise ValidationError("You can't cancel a canceled payment method.")

        cancelable_states = [Transaction.States.Initial,
                             Transaction.States.Pending]

        transactions = self.transactions.filter(state__in=cancelable_states)

        errors = []
        for transaction in transactions:
            if transaction.state == Transaction.States.Initial:
                try:
                    transaction.cancel()
                except TransitionNotAllowed:
                    errors.append("Transaction {} couldn't be canceled".format(transaction.uuid))

            if transaction.state == Transaction.States.Pending:
                payment_processor = self.get_payment_processor()
                if (hasattr(payment_processor, 'void_transaction') and
                        not payment_processor.void_transaction(transaction)):
                    errors.append("Transaction {} couldn't be voided".format(transaction.uuid))

            transaction.save()

        if errors:
            return errors

        self.canceled = True
        self.save()

        return None

    def clean_with_previous_instance(self, previous_instance):
        if not previous_instance:
            return

        for field in self.final_fields:
            old_value = getattr(previous_instance, field, None)
            current_value = getattr(self, field, None)

            if old_value != current_value:
                raise ValidationError(
                    "Field '%s' may not be changed." % field
                )

        for field in self.irreversible_fields:
            old_value = getattr(previous_instance, field, None)
            current_value = getattr(self, field, None)

            if old_value and old_value != current_value:
                raise ValidationError(
                    "Field '%s' may not be changed anymore." % field
                )

    def full_clean(self, *args, **kwargs):
        previous_instance = kwargs.pop('previous_instance', None)

        super(PaymentMethod, self).full_clean(*args, **kwargs)
        self.clean_with_previous_instance(previous_instance)

        # this assumes that nobody calls clean and then modifies this object
        # without calling clean again
        setattr(self, '.cleaned', True)

    @property
    def allowed_currencies(self):
        return self.get_payment_processor().allowed_currencies

    @property
    def public_data(self):
        return {}

    def __str__(self):
        return u'{} - {} - {}'.format(self.customer,
                                      self.get_payment_processor_display(),
                                      self.pk)


def create_transactions_for_issued_documents(payment_method):
    customer = payment_method.customer

    if payment_method.canceled or not payment_method.verified:
        return []

    transactions = []

    for document in chain(
        Proforma.objects.filter(related_document=None, customer=customer,
                                state=Proforma.STATES.ISSUED),
        Invoice.objects.filter(state=Invoice.STATES.ISSUED, customer=customer)
    ):
        try:
            transactions.append(Transaction.objects.create(
                document=document, payment_method=payment_method
            ))
        except ValidationError:
            continue

    return transactions


@receiver(pre_save)
def pre_payment_method_save(sender, instance=None, **kwargs):
    if not isinstance(instance, PaymentMethod):
        return

    payment_method = instance

    previous_instance = get_object_or_None(PaymentMethod, pk=payment_method.pk)
    setattr(payment_method, '.previous_instance', previous_instance)

    if not getattr(payment_method, '.cleaned', False):
        payment_method.full_clean(previous_instance=previous_instance)


@receiver(post_save)
def post_payment_method_save(sender, instance, **kwargs):
    if not isinstance(instance, PaymentMethod):
        return

    payment_method = instance

    if hasattr(payment_method, '.cleaned'):
        delattr(payment_method, '.cleaned')

    previous_instance = getattr(payment_method, '.previous_instance', None)

    if not (settings.SILVER_AUTOMATICALLY_CREATE_TRANSACTIONS or
            not payment_method.verified or
            (not payment_method.get_payment_processor().type ==
                payment_processors.Types.Triggered)):
        return

    if not previous_instance or not previous_instance.verified:
        create_transactions_for_issued_documents(payment_method)



# -*- coding: utf-8 -*-
# Generated by Django 1.11.3 on 2017-09-29 10:45
from __future__ import unicode_literals

from datetime import datetime
from django.db import migrations, models
import django.utils.timezone


def fill_billing_log_created_at_field(apps, schema_editor):
    BillingLog = apps.get_model('silver', 'BillingLog')
    db_alias = schema_editor.connection.alias

    for billing_log in BillingLog.objects.using(db_alias).all():
        billing_log.created_at = datetime.combine(billing_log.billing_date, datetime.min.time())
        billing_log.save(using=db_alias)


class Migration(migrations.Migration):

    dependencies = [
        ('silver', '0040_auto_20170925_0646'),
    ]

    operations = [
        migrations.AddField(
            model_name='billinglog',
            name='created_at',
            field=models.DateTimeField(auto_now_add=True, default=django.utils.timezone.now),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='plan',
            name='cycle_billing_duration',
            field=models.DurationField(blank=True, help_text=b"This can be used to ensure that the billing date doesn't pass a certain date.\nFor example if this field is set to 2 days, for a monthly subscription, the billing date will never surpass the 2nd day of the month. Billing documents can still be generated after that day during the billing cycle, but their billing date will appear to be the end of the cycle billing duration.", null=True),
        ),
        migrations.AddField(
            model_name='provider',
            name='cycle_billing_duration',
            field=models.DurationField(blank=True, help_text=b"This can be used to ensure that the billing date doesn't pass a certain date.\nFor example if this field is set to 2 days, for a monthly subscription, the billing date will never surpass the 2nd day of the month. Billing documents can still be generated after that day during the billing cycle, but their billing date will appear to be the end of the cycle billing duration.", null=True),
        ),
        migrations.RunPython(fill_billing_log_created_at_field, migrations.RunPython.noop)
    ]



# Copyright (c) 2015 Presslabs SRL
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

from django.conf.urls import re_path

from silver import views as silver_views
from silver.api.views import billing_entities_views, documents_views, payment_method_views, \
    plan_views, product_code_views, subscription_views, transaction_views


urlpatterns = [
    re_path(r'^customers/$',
            billing_entities_views.CustomerList.as_view(), name='customer-list'),
    re_path(r'^customers/(?P<customer_pk>[0-9]+)/$',
            billing_entities_views.CustomerDetail.as_view(), name='customer-detail'),

    re_path(r'^customers/(?P<customer_pk>[0-9]+)/subscriptions/$',
            subscription_views.SubscriptionList.as_view(), name='subscription-list'),
    re_path(r'^customers/(?P<customer_pk>[0-9]+)/subscriptions/(?P<subscription_pk>[0-9]+)/$',
            subscription_views.SubscriptionDetail.as_view(), name='subscription-detail'),
    re_path(r'^customers/(?P<customer_pk>[0-9]+)/subscriptions/(?P<subscription_pk>[0-9]+)/metered-features/(?P<mf_product_code>([^/])+)/$',
            subscription_views.MeteredFeatureUnitsLogDetail.as_view(), name='mf-log-units'),
    re_path(r'^customers/(?P<customer_pk>[0-9]+)/subscriptions/(?P<subscription_pk>[0-9]+)/activate/$',
            subscription_views.SubscriptionActivate.as_view(), name='sub-activate'),
    re_path(r'^customers/(?P<customer_pk>[0-9]+)/subscriptions/(?P<subscription_pk>[0-9]+)/cancel/$',
            subscription_views.SubscriptionCancel.as_view(), name='sub-cancel'),
    re_path(r'^customers/(?P<customer_pk>[0-9]+)/subscriptions/(?P<subscription_pk>[0-9]+)/reactivate/$',
            subscription_views.SubscriptionReactivate.as_view(), name='sub-reactivate'),

    re_path(r'^customers/(?P<customer_pk>[0-9]+)/payment_methods/$',
            payment_method_views.PaymentMethodList.as_view(), name='payment-method-list'),
    re_path(r'^customers/(?P<customer_pk>[0-9]+)/payment_methods/(?P<payment_method_id>[0-9]+)/$',
            payment_method_views.PaymentMethodDetail.as_view(), name='payment-method-detail'),
    re_path(r'^customers/(?P<customer_pk>[0-9]+)/payment_methods/(?P<payment_method_id>[0-9]+)/(?P<requested_action>(cancel))_request/$',
            payment_method_views.PaymentMethodAction.as_view(), name='payment-method-action'),

    re_path(r'^customers/(?P<customer_pk>[0-9]+)/payment_methods/(?P<payment_method_id>[0-9]+)/transactions/$',
            transaction_views.TransactionList.as_view(), name='payment-method-transaction-list'),
    re_path(r'^customers/(?P<customer_pk>[0-9]+)/transactions/$',
            transaction_views.TransactionList.as_view(), name='transaction-list'),
    re_path(r'^customers/(?P<customer_pk>[0-9]+)/transactions/(?P<transaction_uuid>[0-9a-z-]+)/$',
            transaction_views.TransactionDetail.as_view(), name='transaction-detail'),
    re_path(r'^customers/(?P<customer_pk>[0-9]+)/transactions/(?P<transaction_uuid>[0-9a-z-]+)/(?P<requested_action>(cancel))_request/$',
            transaction_views.TransactionAction.as_view(), name='transaction-action'),

    re_path(r'^payment_processors/$',
            payment_method_views.PaymentProcessorList.as_view(), name='payment-processor-list'),
    re_path(r'^payment_processors/(?P<processor_name>[a-zA-Z\-\_]+)/$',
            payment_method_views.PaymentProcessorDetail.as_view(), name='payment-processor-detail'),

    re_path(r'^plans/$',
            plan_views.PlanList.as_view(), name='plan-list'),
    re_path(r'^plans/(?P<pk>[0-9]+)/$',
            plan_views.PlanDetail.as_view(), name='plan-detail'),
    re_path(r'plans/(?P<pk>[0-9]+)/metered-features/$',
            plan_views.PlanMeteredFeatures.as_view(), name='plans-metered-features'),

    re_path(r'^metered-features/$',
            subscription_views.MeteredFeatureList.as_view(), name='metered-feature-list'),

    re_path(r'^providers/$',
            billing_entities_views.ProviderListCreate.as_view(), name='provider-list'),
    re_path(r'^providers/(?P<pk>[0-9]+)/$',
            billing_entities_views.ProviderRetrieveUpdateDestroy.as_view(), name='provider-detail'),

    re_path(r'^product-codes/$',
            product_code_views.ProductCodeListCreate.as_view(), name='productcode-list'),
    re_path(r'^product-codes/(?P<pk>[0-9]+)/$',
            product_code_views.ProductCodeRetrieveUpdate.as_view(), name='productcode-detail'),

    re_path(r'^invoices/$',
            documents_views.InvoiceListCreate.as_view(), name='invoice-list'),
    re_path(r'^invoices/(?P<pk>[0-9]+)/$',
            documents_views.InvoiceRetrieveUpdate.as_view(), name='invoice-detail'),
    re_path(r'^invoices/(?P<document_pk>[0-9]+)/entries/$',
            documents_views.InvoiceEntryCreate.as_view(), name='invoice-entry-create'),
    re_path(r'^invoices/(?P<document_pk>[0-9]+)/entries/(?P<entry_pk>[0-9]+)/$',
            documents_views.InvoiceEntryUpdateDestroy.as_view(), name='invoice-entry-update'),
    re_path(r'^invoices/(?P<pk>[0-9]+)/state/$',
            documents_views.InvoiceStateHandler.as_view(), name='invoice-state'),
    re_path(r'^invoices/(?P<invoice_id>\d+).pdf$',
            silver_views.invoice_pdf, name='invoice-pdf'),

    re_path(r'^proformas/$',
            documents_views.ProformaListCreate.as_view(), name='proforma-list'),
    re_path(r'^proformas/(?P<pk>[0-9]+)/$',
            documents_views.ProformaRetrieveUpdate.as_view(), name='proforma-detail'),
    re_path(r'^proformas/(?P<document_pk>[0-9]+)/entries/$',
            documents_views.ProformaEntryCreate.as_view(), name='proforma-entry-create'),
    re_path(r'^proformas/(?P<document_pk>[0-9]+)/entries/(?P<entry_pk>[0-9]+)/$',
            documents_views.ProformaEntryUpdateDestroy.as_view(),
            name='proforma-entry-update'),
    re_path(r'^proformas/(?P<pk>[0-9]+)/state/$',
            documents_views.ProformaStateHandler.as_view(), name='proforma-state'),
    re_path(r'^proformas/(?P<pk>[0-9]+)/invoice/$',
            documents_views.ProformaInvoiceRetrieveCreate.as_view(),
            name='proforma-invoice'),
    re_path(r'^proformas/(?P<proforma_id>\d+).pdf$',
            silver_views.proforma_pdf, name='proforma-pdf'),
    re_path(r'^pdfs/(?P<pdf_pk>[0-9]+)/$',
            documents_views.PDFRetrieve.as_view(),
            name='pdf'),
    re_path(r'^documents/$',
            documents_views.DocumentList.as_view(), name='document-list')
]
